@@ -2,7 +2,7 @@
 
   class.c -
 
-  $Author: usa $
+  $Author: naruse $
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -274,7 +274,7 @@ clone_method_i(ID key, VALUE value, void *data)
 
 struct clone_const_arg {
     VALUE klass;
-    st_table *tbl;
+    struct rb_id_table *tbl;
 };
 
 static int
@@ -285,14 +285,14 @@ clone_const(ID key, const rb_const_entry_t *ce, struct clone_const_arg *arg)
     RB_OBJ_WRITTEN(arg->klass, Qundef, ce->value);
     RB_OBJ_WRITTEN(arg->klass, Qundef, ce->file);
 
-    st_insert(arg->tbl, key, (st_data_t)nce);
-    return ST_CONTINUE;
+    rb_id_table_insert(arg->tbl, key, (VALUE)nce);
+    return ID_TABLE_CONTINUE;
 }
 
-static int
-clone_const_i(st_data_t key, st_data_t value, st_data_t data)
+static enum rb_id_table_iterator_result
+clone_const_i(ID key, VALUE value, void *data)
 {
-    return clone_const((ID)key, (const rb_const_entry_t *)value, (struct clone_const_arg *)data);
+    return clone_const(key, (const rb_const_entry_t *)value, data);
 }
 
 static void
@@ -346,10 +346,9 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
     if (RCLASS_CONST_TBL(orig)) {
 	struct clone_const_arg arg;
 
-	RCLASS_CONST_TBL(clone) = st_init_numtable();
+	arg.tbl = RCLASS_CONST_TBL(clone) = rb_id_table_create(0);
 	arg.klass = clone;
-	arg.tbl = RCLASS_CONST_TBL(clone);
-	st_foreach(RCLASS_CONST_TBL(orig), clone_const_i, (st_data_t)&arg);
+	rb_id_table_foreach(RCLASS_CONST_TBL(orig), clone_const_i, &arg);
     }
     if (RCLASS_M_TBL(orig)) {
 	struct clone_method_arg arg;
@@ -393,10 +392,9 @@ rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 	}
 	if (RCLASS_CONST_TBL(klass)) {
 	    struct clone_const_arg arg;
-	    RCLASS_CONST_TBL(clone) = st_init_numtable();
+	    arg.tbl = RCLASS_CONST_TBL(clone) = rb_id_table_create(0);
 	    arg.klass = clone;
-	    arg.tbl = RCLASS_CONST_TBL(clone);
-	    st_foreach(RCLASS_CONST_TBL(klass), clone_const_i, (st_data_t)&arg);
+	    rb_id_table_foreach(RCLASS_CONST_TBL(klass), clone_const_i, &arg);
 	}
 	if (attach != Qundef) {
 	    rb_singleton_class_attached(clone, attach);
@@ -550,6 +548,10 @@ Init_class_hierarchy(void)
     rb_cBasicObject = boot_defclass("BasicObject", 0);
     rb_cObject = boot_defclass("Object", rb_cBasicObject);
     rb_gc_register_mark_object(rb_cObject);
+
+    /* resolve class name ASAP for order-independence */
+    rb_class_name(rb_cObject);
+
     rb_cModule = boot_defclass("Module", rb_cObject);
     rb_cClass =  boot_defclass("Class",  rb_cModule);
 
@@ -629,12 +631,12 @@ rb_class_inherited(VALUE super, VALUE klass)
  * Defines a top-level class.
  * \param name   name of the class
  * \param super  a class from which the new class will derive.
- *               NULL means \c Object class.
  * \return the created class
  * \throw TypeError if the constant name \a name is already taken but
  *                  the constant is not a \c Class.
- * \throw NameError if the class is already defined but the class can not
+ * \throw TypeError if the class is already defined but the class can not
  *                  be reopened because its superclass is not \a super.
+ * \throw ArgumentError if the \a super is NULL.
  * \post top-level constant named \a name refers the returned class.
  *
  * \note if a class named \a name is already defined and its superclass is
@@ -659,7 +661,7 @@ rb_define_class(const char *name, VALUE super)
 	return klass;
     }
     if (!super) {
-	rb_warn("no super class for `%s', Object assumed", name);
+	rb_raise(rb_eArgError, "no super class for `%s'", name);
     }
     klass = rb_define_class_id(id, super);
     rb_vm_add_root_module(id, klass);
@@ -731,8 +733,8 @@ rb_define_class_id_under(VALUE outer, ID id, VALUE super)
 	return klass;
     }
     if (!super) {
-	rb_warn("no super class for `%"PRIsVALUE"::%"PRIsVALUE"', Object assumed",
-		rb_class_path(outer), rb_id2str(id));
+	rb_raise(rb_eArgError, "no super class for `%"PRIsVALUE"::%"PRIsVALUE"'",
+		 rb_class_path(outer), rb_id2str(id));
     }
     klass = rb_define_class_id(id, super);
     rb_set_class_path_string(klass, outer, rb_id2str(id));
@@ -824,7 +826,7 @@ rb_include_class_new(VALUE module, VALUE super)
 	RCLASS_IV_TBL(module) = st_init_numtable();
     }
     if (!RCLASS_CONST_TBL(module)) {
-	RCLASS_CONST_TBL(module) = st_init_numtable();
+	RCLASS_CONST_TBL(module) = rb_id_table_create(0);
     }
     RCLASS_IV_TBL(klass) = RCLASS_IV_TBL(module);
     RCLASS_CONST_TBL(klass) = RCLASS_CONST_TBL(module);
@@ -847,23 +849,14 @@ rb_include_class_new(VALUE module, VALUE super)
 
 static int include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super);
 
-static void
-ensure_includable(VALUE klass, VALUE module)
-{
-    rb_frozen_class_p(klass);
-    Check_Type(module, T_MODULE);
-    if (!NIL_P(rb_refinement_module_get_refined_class(module))) {
-	rb_raise(rb_eArgError, "refinement module is not allowed");
-    }
-    OBJ_INFECT(klass, module);
-}
-
 void
 rb_include_module(VALUE klass, VALUE module)
 {
     int changed = 0;
 
-    ensure_includable(klass, module);
+    rb_frozen_class_p(klass);
+    Check_Type(module, T_MODULE);
+    OBJ_INFECT(klass, module);
 
     changed = include_modules_at(klass, RCLASS_ORIGIN(klass), module, TRUE);
     if (changed < 0)
@@ -886,6 +879,7 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
 
     while (module) {
 	int superclass_seen = FALSE;
+	struct rb_id_table *tbl;
 
 	if (RCLASS_ORIGIN(module) != module)
 	    goto skip;
@@ -923,8 +917,12 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
 	    rb_id_table_foreach(RMODULE_M_TBL(module), add_refined_method_entry_i, (void *)refined_class);
 	    FL_SET(c, RMODULE_INCLUDED_INTO_REFINEMENT);
 	}
-	if (RMODULE_M_TBL(module) && rb_id_table_size(RMODULE_M_TBL(module))) method_changed = 1;
-	if (RMODULE_CONST_TBL(module) && RMODULE_CONST_TBL(module)->num_entries) constant_changed = 1;
+
+	tbl = RMODULE_M_TBL(module);
+	if (tbl && rb_id_table_size(tbl)) method_changed = 1;
+
+	tbl = RMODULE_CONST_TBL(module);
+	if (tbl && rb_id_table_size(tbl)) constant_changed = 1;
       skip:
 	module = RCLASS_SUPER(module);
     }
@@ -968,7 +966,9 @@ rb_prepend_module(VALUE klass, VALUE module)
     VALUE origin;
     int changed = 0;
 
-    ensure_includable(klass, module);
+    rb_frozen_class_p(klass);
+    Check_Type(module, T_MODULE);
+    OBJ_INFECT(klass, module);
 
     origin = RCLASS_ORIGIN(klass);
     if (origin == klass) {
@@ -1090,46 +1090,51 @@ rb_mod_ancestors(VALUE mod)
     return ary;
 }
 
-static int
-ins_methods_push(ID name, rb_method_visibility_t visi, VALUE ary, rb_method_visibility_t expected_visi)
+static void
+ins_methods_push(st_data_t name, st_data_t ary)
 {
-    if (visi == METHOD_VISI_UNDEF) return ST_CONTINUE;
+    rb_ary_push((VALUE)ary, ID2SYM((ID)name));
+}
 
-    switch (expected_visi) {
+static int
+ins_methods_i(st_data_t name, st_data_t type, st_data_t ary)
+{
+    switch ((rb_method_visibility_t)type) {
       case METHOD_VISI_UNDEF:
-	if (visi != METHOD_VISI_PRIVATE) rb_ary_push(ary, ID2SYM(name));
-	break;
       case METHOD_VISI_PRIVATE:
-      case METHOD_VISI_PROTECTED:
-      case METHOD_VISI_PUBLIC:
-	if (visi == expected_visi) rb_ary_push(ary, ID2SYM(name));
+	break;
+      default: /* everything but private */
+	ins_methods_push(name, ary);
 	break;
     }
     return ST_CONTINUE;
 }
 
-static int
-ins_methods_i(st_data_t name, st_data_t type, st_data_t ary)
-{
-    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_UNDEF); /* everything but private */
-}
-
 static int
 ins_methods_prot_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PROTECTED);
+    if ((rb_method_visibility_t)type == METHOD_VISI_PROTECTED) {
+	ins_methods_push(name, ary);
+    }
+    return ST_CONTINUE;
 }
 
 static int
 ins_methods_priv_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PRIVATE);
+    if ((rb_method_visibility_t)type == METHOD_VISI_PRIVATE) {
+	ins_methods_push(name, ary);
+    }
+    return ST_CONTINUE;
 }
 
 static int
 ins_methods_pub_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PUBLIC);
+    if ((rb_method_visibility_t)type == METHOD_VISI_PUBLIC) {
+	ins_methods_push(name, ary);
+    }
+    return ST_CONTINUE;
 }
 
 struct method_entry_arg {
@@ -1523,6 +1528,23 @@ rb_undef_method(VALUE klass, const char *name)
     rb_add_method(klass, rb_intern(name), VM_METHOD_TYPE_UNDEF, 0, METHOD_VISI_UNDEF);
 }
 
+static enum rb_id_table_iterator_result
+undef_method_i(ID name, VALUE value, void *data)
+{
+    VALUE klass = (VALUE)data;
+    rb_add_method(klass, name, VM_METHOD_TYPE_UNDEF, 0, METHOD_VISI_UNDEF);
+    return ID_TABLE_CONTINUE;
+}
+
+void
+rb_undef_methods_from(VALUE klass, VALUE super)
+{
+    struct rb_id_table *mtbl = RCLASS_M_TBL(super);
+    if (mtbl) {
+	rb_id_table_foreach(mtbl, undef_method_i, (void *)klass);
+    }
+}
+
 /*!
  * \}
  */
@@ -1589,7 +1611,9 @@ singleton_class_of(VALUE obj)
     klass = RBASIC(obj)->klass;
     if (!(FL_TEST(klass, FL_SINGLETON) &&
 	  rb_ivar_get(klass, id_attached) == obj)) {
+	rb_serial_t serial = RCLASS_SERIAL(klass);
 	klass = rb_make_metaclass(obj, klass);
+	RCLASS_SERIAL(klass) = serial;
     }
 
     if (OBJ_TAINTED(obj)) {
@@ -1598,7 +1622,7 @@ singleton_class_of(VALUE obj)
     else {
 	FL_UNSET(klass, FL_TAINT);
     }
-    if (OBJ_FROZEN(obj)) OBJ_FREEZE_RAW(klass);
+    RB_FL_SET_RAW(klass, RB_OBJ_FROZEN_RAW(obj));
 
     return klass;
 }
@@ -1641,7 +1665,7 @@ rb_singleton_class_get(VALUE obj)
  * Returns the singleton class of \a obj. Creates it if necessary.
  *
  * \param obj an arbitrary object.
- * \throw TypeError if \a obj is a Fixnum or a Symbol.
+ * \throw TypeError if \a obj is a Integer or a Symbol.
  * \return the singleton class.
  *
  * \post \a obj has its own singleton class.
@@ -1752,141 +1776,6 @@ rb_obj_basic_to_s_p(VALUE obj)
     return 0;
 }
 
-#include <stdarg.h>
-
-int
-rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
-{
-    int i;
-    const char *p = fmt;
-    VALUE *var;
-    va_list vargs;
-    int f_var = 0, f_hash = 0, f_block = 0;
-    int n_lead = 0, n_opt = 0, n_trail = 0, n_mand;
-    int argi = 0;
-    VALUE hash = Qnil;
-
-    if (ISDIGIT(*p)) {
-	n_lead = *p - '0';
-	p++;
-	if (ISDIGIT(*p)) {
-	    n_opt = *p - '0';
-	    p++;
-	    if (ISDIGIT(*p)) {
-		n_trail = *p - '0';
-		p++;
-		goto block_arg;
-	    }
-	}
-    }
-    if (*p == '*') {
-	f_var = 1;
-	p++;
-	if (ISDIGIT(*p)) {
-	    n_trail = *p - '0';
-	    p++;
-	}
-    }
-  block_arg:
-    if (*p == ':') {
-	f_hash = 1;
-	p++;
-    }
-    if (*p == '&') {
-	f_block = 1;
-	p++;
-    }
-    if (*p != '\0') {
-	rb_fatal("bad scan arg format: %s", fmt);
-    }
-    n_mand = n_lead + n_trail;
-
-    if (argc < n_mand)
-	goto argc_error;
-
-    va_start(vargs, fmt);
-
-    /* capture an option hash - phase 1: pop */
-    if (f_hash && n_mand < argc) {
-	VALUE last = argv[argc - 1];
-
-	if (NIL_P(last)) {
-	    /* nil is taken as an empty option hash only if it is not
-	       ambiguous; i.e. '*' is not specified and arguments are
-	       given more than sufficient */
-	    if (!f_var && n_mand + n_opt < argc)
-		argc--;
-	}
-	else {
-	    hash = rb_check_hash_type(last);
-	    if (!NIL_P(hash)) {
-		VALUE opts = rb_extract_keywords(&hash);
-		if (!hash) argc--;
-		hash = opts ? opts : Qnil;
-	    }
-	}
-    }
-    /* capture leading mandatory arguments */
-    for (i = n_lead; i-- > 0; ) {
-	var = va_arg(vargs, VALUE *);
-	if (var) *var = argv[argi];
-	argi++;
-    }
-    /* capture optional arguments */
-    for (i = n_opt; i-- > 0; ) {
-	var = va_arg(vargs, VALUE *);
-	if (argi < argc - n_trail) {
-	    if (var) *var = argv[argi];
-	    argi++;
-	}
-	else {
-	    if (var) *var = Qnil;
-	}
-    }
-    /* capture variable length arguments */
-    if (f_var) {
-	int n_var = argc - argi - n_trail;
-
-	var = va_arg(vargs, VALUE *);
-	if (0 < n_var) {
-	    if (var) *var = rb_ary_new4(n_var, &argv[argi]);
-	    argi += n_var;
-	}
-	else {
-	    if (var) *var = rb_ary_new();
-	}
-    }
-    /* capture trailing mandatory arguments */
-    for (i = n_trail; i-- > 0; ) {
-	var = va_arg(vargs, VALUE *);
-	if (var) *var = argv[argi];
-	argi++;
-    }
-    /* capture an option hash - phase 2: assignment */
-    if (f_hash) {
-	var = va_arg(vargs, VALUE *);
-	if (var) *var = hash;
-    }
-    /* capture iterator block */
-    if (f_block) {
-	var = va_arg(vargs, VALUE *);
-	if (rb_block_given_p()) {
-	    *var = rb_block_proc();
-	}
-	else {
-	    *var = Qnil;
-	}
-    }
-    va_end(vargs);
-
-    if (argi < argc) {
-      argc_error:
-	rb_error_arity(argc, n_mand, f_var ? UNLIMITED_ARGUMENTS : n_mand + n_opt);
-    }
-
-    return argc;
-}
-
 VALUE
 rb_keyword_error_new(const char *error, VALUE keys)
 {
@@ -2018,6 +1907,134 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 #undef extract_kwarg
 }
 
+#undef rb_scan_args
+int
+rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
+{
+    int i;
+    const char *p = fmt;
+    VALUE *var;
+    va_list vargs;
+    int f_var = 0, f_hash = 0, f_block = 0;
+    int n_lead = 0, n_opt = 0, n_trail = 0, n_mand;
+    int argi = 0;
+    VALUE hash = Qnil;
+
+    if (ISDIGIT(*p)) {
+	n_lead = *p - '0';
+	p++;
+	if (ISDIGIT(*p)) {
+	    n_opt = *p - '0';
+	    p++;
+	}
+    }
+    if (*p == '*') {
+	f_var = 1;
+	p++;
+    }
+    if (ISDIGIT(*p)) {
+	n_trail = *p - '0';
+	p++;
+    }
+    if (*p == ':') {
+	f_hash = 1;
+	p++;
+    }
+    if (*p == '&') {
+	f_block = 1;
+	p++;
+    }
+    if (*p != '\0') {
+	rb_fatal("bad scan arg format: %s", fmt);
+    }
+    n_mand = n_lead + n_trail;
+
+    if (argc < n_mand)
+	goto argc_error;
+
+    va_start(vargs, fmt);
+
+    /* capture an option hash - phase 1: pop */
+    if (f_hash && n_mand < argc) {
+	VALUE last = argv[argc - 1];
+
+	if (NIL_P(last)) {
+	    /* nil is taken as an empty option hash only if it is not
+	       ambiguous; i.e. '*' is not specified and arguments are
+	       given more than sufficient */
+	    if (!f_var && n_mand + n_opt < argc)
+		argc--;
+	}
+	else {
+	    hash = rb_check_hash_type(last);
+	    if (!NIL_P(hash)) {
+		VALUE opts = rb_extract_keywords(&hash);
+		if (!hash) argc--;
+		hash = opts ? opts : Qnil;
+	    }
+	}
+    }
+    /* capture leading mandatory arguments */
+    for (i = n_lead; i-- > 0; ) {
+	var = va_arg(vargs, VALUE *);
+	if (var) *var = argv[argi];
+	argi++;
+    }
+    /* capture optional arguments */
+    for (i = n_opt; i-- > 0; ) {
+	var = va_arg(vargs, VALUE *);
+	if (argi < argc - n_trail) {
+	    if (var) *var = argv[argi];
+	    argi++;
+	}
+	else {
+	    if (var) *var = Qnil;
+	}
+    }
+    /* capture variable length arguments */
+    if (f_var) {
+	int n_var = argc - argi - n_trail;
+
+	var = va_arg(vargs, VALUE *);
+	if (0 < n_var) {
+	    if (var) *var = rb_ary_new4(n_var, &argv[argi]);
+	    argi += n_var;
+	}
+	else {
+	    if (var) *var = rb_ary_new();
+	}
+    }
+    /* capture trailing mandatory arguments */
+    for (i = n_trail; i-- > 0; ) {
+	var = va_arg(vargs, VALUE *);
+	if (var) *var = argv[argi];
+	argi++;
+    }
+    /* capture an option hash - phase 2: assignment */
+    if (f_hash) {
+	var = va_arg(vargs, VALUE *);
+	if (var) *var = hash;
+    }
+    /* capture iterator block */
+    if (f_block) {
+	var = va_arg(vargs, VALUE *);
+	if (rb_block_given_p()) {
+	    *var = rb_block_proc();
+	}
+	else {
+	    *var = Qnil;
+	}
+    }
+    va_end(vargs);
+
+    if (argi < argc) {
+      argc_error:
+	rb_error_arity(argc, n_mand, f_var ? UNLIMITED_ARGUMENTS : n_mand + n_opt);
+    }
+
+    return argc;
+}
+
 int
 rb_class_has_methods(VALUE c)
 {
