@@ -2,7 +2,7 @@
 
   main.c -
 
-  $Author$
+  $Author: ko1 $
   created at: Fri Aug 19 13:19:58 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
